@@ -7,7 +7,18 @@ import re
 import warnings
 from functools import lru_cache, partial, wraps
 from hashlib import md5
-from typing import Any, Callable, Dict, Iterable, Iterator, Tuple, Union, cast, overload
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    Iterator,
+    NamedTuple,
+    Tuple,
+    Union,
+    cast,
+    overload,
+)
 
 import jax
 import jax.numpy as jnp
@@ -16,6 +27,7 @@ import jax.scipy.linalg
 import numpy as np
 import orjson
 from natsort import natsorted
+from numpy.exceptions import ComplexWarning
 
 from .saxtypes import (
     ComplexArrayND,
@@ -57,18 +69,12 @@ def block_diag(*arrs: ComplexArrayND) -> ComplexArrayND:
 
 def clean_string(s: str, dot="p", minus="m", other="_") -> str:
     """clean a string such that it is a valid python identifier"""
-    original = s
     s = s.strip()
     s = s.replace(".", dot)  # dot
     s = s.replace("-", minus)  # minus
-    s = re.sub("[^0-9a-zA-Z]", other, s)
+    s = re.sub("[^0-9a-zA-Z_]", other, s)
     if s[0] in "0123456789":
         s = "_" + s
-    if s != original:
-        warnings.warn(
-            f"modified string {original} in an attempt "
-            f"to make valid python identifier: {s}"
-        )
     if not s.isidentifier():
         raise ValueError(f"failed to clean string to a valid python identifier: {s}")
     return s
@@ -93,10 +99,10 @@ def validate_settings(settings: Settings) -> Settings:
 def try_complex_float(f: Any) -> Any:
     """try converting an object to float, return unchanged object on fail"""
     with warnings.catch_warnings():
-        warnings.filterwarnings(action="error", category=jnp.ComplexWarning)
+        warnings.filterwarnings(action="error", category=ComplexWarning)
         try:
             return jnp.asarray(f, dtype=float)
-        except jnp.ComplexWarning:  # type: ignore
+        except ComplexWarning:
             return jnp.asarray(f, dtype=complex)
         except (ValueError, TypeError):
             return f
@@ -598,3 +604,38 @@ def _numpyfy(obj: Any):
         return np.asarray(obj)
     else:
         return {k: _numpyfy(v) for k, v in obj.items()}
+
+
+class Normalization(NamedTuple):
+    mean: ComplexArrayND
+    std: ComplexArrayND
+
+
+def normalization(x: ComplexArrayND, axis=None):
+    if axis is None:
+        return Normalization(x.mean(), x.std())
+    else:
+        return Normalization(x.mean(axis), x.std(axis))
+
+
+def cartesian_product(*arrays: ComplexArrayND) -> ComplexArrayND:
+    """calculate the n-dimensional cartesian product of an arbitrary number of arrays"""
+    ixarrays = jnp.ix_(*arrays)
+    barrays = jnp.broadcast_arrays(*ixarrays)
+    sarrays = jnp.stack(barrays, -1)
+    assert isinstance(sarrays, jnp.ndarray)
+    product = sarrays.reshape(-1, sarrays.shape[-1])
+    assert isinstance(product, jnp.ndarray)
+    return product
+
+
+def normalize(x: ComplexArrayND, normalization: Normalization) -> Tuple[ComplexArrayND]:
+    """normalize an array with a given mean and standard deviation"""
+    mean, std = normalization
+    return (x - mean) / std
+
+
+def denormalize(x: ComplexArrayND, normalization: Normalization) -> ComplexArrayND:
+    """denormalize an array with a given mean and standard deviation"""
+    mean, std = normalization
+    return x * std + mean
